/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.combine;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Random;

/**
 * @author Henrik Kaipe
 */
class AllPossible2Combinations {

    interface Picker {
        void pick(Object[] record);
        boolean all2CombinationsAreSatisfied();
    }
    
    private static class PickerFactory {
        private final Object[][] variableValues;
        private final int lowestPossibleNumberOfRecords;

        private final Random random;

        PickerFactory(Object[][] variableValues, Random random) {
            this.variableValues = variableValues;
            this.random = random;
            
            lowestPossibleNumberOfRecords =
                0 == variableValues.length ? 1 :
                1 == variableValues.length ? variableValues[0].length :
                variableValues[0].length * variableValues[1].length;
        }
        
        Picker newPicker(final int pickerIndex) {
            
            abstract class DeterministicCountBasedPicker implements Picker {
                final int nbrOfOptions = variableValues[pickerIndex].length;
                
                private int count = 0;

                abstract int chooseValuesIndex(int count);

                public void pick(final Object[] record) {
                    record[pickerIndex] = variableValues
                            [pickerIndex][chooseValuesIndex(count)];
                    ++count;
                }
                public boolean all2CombinationsAreSatisfied() {
                    return lowestPossibleNumberOfRecords <= count;
                }
            }
            
            switch (pickerIndex) {
                case 0:
                    return new DeterministicCountBasedPicker() {
                        int chooseValuesIndex(final int count) {
                            return count % nbrOfOptions;
                        }
                    };
                case 1:
                    return new DeterministicCountBasedPicker() {
                        final int nbrOfObjectsToCombineWith =
                                variableValues[0].length;
                        int chooseValuesIndex(final int count) {
                            return (
                                count % nbrOfObjectsToCombineWith
                                + count / nbrOfObjectsToCombineWith
                            ) % nbrOfOptions;
                        }
                    };
                case 2:
                    return new DeterministicCountBasedPicker() {
                        final int nbrOfCombineValuesAt0 =
                                variableValues[0].length;
                        final int nbrOfCombineValuesAt1 =
                                variableValues[1].length;

                        int chooseValuesIndex(int count) {
                            return (
                                + count
                                / nbrOfCombineValuesAt0 / nbrOfCombineValuesAt1
                                + count
                                / nbrOfCombineValuesAt0
                                + count
                                % nbrOfCombineValuesAt0 / nbrOfCombineValuesAt1
                            ) % nbrOfOptions;
                        }
                    };
                default:
                    return new Default2CombinationPicker(
                            pickerIndex, variableValues, random);
            }
        }
    }
    
    /**
     * This is where the real work is performed. Stateful pickers are
     * created from the variableValues specified.
     * The pick-method of picker[x] will pick its values from the
     * values available in the array found at variableValues[x].
     */
    private static Picker[] createPickers(
            final Object[][] variableValues, final Random random) {
        Picker[] p = new Picker[variableValues.length];

        final PickerFactory pf = new PickerFactory(variableValues, random);
        for (int i = 0 ; i < p.length ; ++i) {
            p[i] = pf.newPicker(i);
        }
        return p;
    }
    
    /**
     * By using this factory-method and specify a certain seed -
     * it is possible to make sure the same combinations
     * will reappear on each test-run.
     */
    static Collection combine(ValuesCollection vc, long randomSeed) {
        
        final Object[] recordTemplate = new Object[vc.size()];
        for (int i = 0, length = vc.size() ; i < length ; ++i) {
            final Object[] values = vc.get(i);
            if (1 == values.length) {
                recordTemplate[i] = values[0];
            }
        }
        
        final Integer[] orderOfVcIndexes = indexWithMostValuesFirst(vc);
        if (0 == orderOfVcIndexes.length) {
            return Arrays.asList(new Object[][] {recordTemplate});
        }
        
        final Object[][] variableValues = new Object[orderOfVcIndexes.length][];
        for (int i = 0 ; i < orderOfVcIndexes.length ; ++i) {
            variableValues[i] = vc.get(orderOfVcIndexes[i].intValue());
        }

        final Picker[] pickers = createPickers(
                variableValues, new Random(randomSeed));
        
        final List recordList = new ArrayList();
        final Object[] pickRecord = new Object[pickers.length];
        for (int satisfyIx = 0 ; satisfyIx < pickers.length ; ++satisfyIx) {
            final Picker nextPickerToSatisfy = pickers[satisfyIx];
            
            while (false == nextPickerToSatisfy.all2CombinationsAreSatisfied()) {
                for (int pickIx = 0 ; pickIx < pickers.length ; ++pickIx) {
                    pickers[pickIx].pick(pickRecord);
                }
                final Object[] nextRecord = (Object[]) recordTemplate.clone();
                for (int i = 0 ; i < pickRecord.length ; ++i) {
                    nextRecord[orderOfVcIndexes[i].intValue()] = pickRecord[i];
                }
                recordList.add(nextRecord);
            }
        }
        return recordList;
    }

    private static Integer[] indexWithMostValuesFirst(final ValuesCollection vc) {
        final List mostValuesFirst = new ArrayList(vc.size());
        
        for (int i = 0, length = vc.size() ; i < length ; ++i) {
            if (2 <= vc.get(i).length) {
                mostValuesFirst.add(new Integer(i));
            }
        }
        
        Collections.sort(mostValuesFirst, new Comparator() {
            public int compare(final Object o1, final Object o2) {
                final Integer i1 = (Integer) o1;
                final Integer i2 = (Integer) o2;
                
                final int sizeDiff = vc.get(i1.intValue()).length
                        - vc.get(i2.intValue()).length;
                if (0 != sizeDiff) {
                    return - sizeDiff;
                } else {
                    return i1.compareTo(i2);
                }
            }
        });
        
        return (Integer[])
                mostValuesFirst.toArray(new Integer[mostValuesFirst.size()]);
    }
}
