/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.combine.reflect;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;
import org.callbackparams.combine.CombineAllPossible2Combinations;
import org.callbackparams.combine.CombineStrategy;
import org.callbackparams.combine.ValuesCollection;

/**
 * Factory-class whose main purpose is to create callback-records from
 * reflectively looked-up callback-values factory classes. The default
 * recipe for a callback-values factory is a nested class that has a
 * static array-returning method with the signature 'values()', e.g. a nested
 * enum.
 * @author Henrik Kaipe
 */
public abstract class CallbackRecordsFactory {
    
    private static CallbackRecordsFactory instance;

    /**
     * Factory method that returns the implementation suitable for the
     * java-version in use. I.e. a pretty simple implementation for JDK-1.4,
     * which basicly only supports a default behaviour, and a more feature-rich
     * implementation for JDK-1.5+, which makes it possible tweak the default
     * behaviour through annotations.
     */
    public static CallbackRecordsFactory getInstance() {
        if (null == instance) {
            String annotationAwareFactoryClassName =
                    "org.callbackparams.combine.annotation.AnnotationAwareCallbackRecordsFactory";
            try {
                instance = (CallbackRecordsFactory)
                        Class.forName(annotationAwareFactoryClassName)
                        .newInstance();
//                System.out.println("CallbackRecordsFactory with TIGER support");

            } catch (UnsupportedClassVersionError detectionOf_JDK1_4) {
                instance = new CallbackRecordsFactory() {};
//                System.out.println("CallbackRecordsFactory for JDK-1.4");

            } catch (Exception shouldNeverHappen) {
                throw new Error(shouldNeverHappen);
            }
        }
        return instance;
    }
    
    /**
     * Searches the specified class, its super-classes and all implemented
     * interfaces for nested classes that provide callback-values, as
     * defined by {@link #isCallbackValuesFactoryClass(java.lang.Class)}.
     */
    protected Set findCallbackValuesFactoryClasses(Class testClass) {
        if (null == testClass) {
            return new LinkedHashSet();
        }
        final Set factoryClasses =
                findCallbackValuesFactoryClasses(testClass.getSuperclass());
        
        final Class[] interfaces = testClass.getInterfaces();
        for (int i = 0 ; i < interfaces.length ; ++i) {
            factoryClasses.addAll(
                    findCallbackValuesFactoryClasses(interfaces[i]));
        }
        
        final Class[] nestedClasses = testClass.getDeclaredClasses();
        for (int i = 0 ; i < nestedClasses.length ; ++i) {
            if (isCallbackValuesFactoryClass(nestedClasses[i])) {
                factoryClasses.add(nestedClasses[i]);
            }
        }
        return factoryClasses;
    }
    
    /**
     * The default implementation returns true if the specified class has a
     * static method "values()" that returns an array.
     */
    protected boolean isCallbackValuesFactoryClass(Class factoryProspect) {
        try {
            Method factoryMethod = factoryProspect
                    .getDeclaredMethod("values", (Class[]) null);
            return 0 < (factoryMethod.getModifiers() & Modifier.STATIC)
                    && factoryMethod.getReturnType().isArray();
            
        } catch (NoSuchMethodException x) {
            return false;
        }
    }
    
    /**
     * The returned array (supposedly containing callback-values) is
     * determined from the specified class. The default procedure to determine
     * this array is simply to invoke the static method values() and
     * assume it returns an array of callback-values.
     * 
     * The reason is that the static values()-method is the method that returns
     * an array of all enum-constants for a java.lang.Enum implementation.
     * <br/>
     * However, it is not necessary that the argument class is an Enum, as long
     * as it has its static values()-method returning an array of
     * callback-values.
     */
    public Object[] retrieveCallbackValuesArray(
            Class callbackValuesFactoryClass) {
        try {
            Method m = callbackValuesFactoryClass
                    .getDeclaredMethod("values", (Class[]) null);
            if (false == m.isAccessible()) {
                m.setAccessible(true);
            }
            return (Object[]) m.invoke(null, (Object[])null);
            
        } catch (Exception x) {
            throw new Error(x);
        }
    }

    /**
     * The default combine strategy is {@link CombineAllPossible2Combinations}
     * but any other strategy could be used by overriding
     * this class or by using the
     * {@link org.callbackparams.combine.annotation.CombineConfig @CombineConfig}
     * annotation (if using JDK-1.5+).
     */
    public CombineStrategy retrieveCombineStrategy(Class testClass) {
        return new CombineAllPossible2Combinations();
    }

    /**
     * This is the method that integration harnesses (such as the
     * {@link org.callbackparams.junit4.CallbackParamsRunner CallbackParamsRunner})
     * should use for retrieving their callback-records in a test-developer
     * friendly way.
     *
     * @param testClass the test class
     * @return the collection of combined and ready-to-use callback-records
     *         retrieved from testClass
     */
    public Collection collectCallbackRecordsReflectively(Class testClass) {
        ValuesCollection vc = new ValuesCollection();
        for (Iterator i = findCallbackValuesFactoryClasses(testClass)
                .iterator() ; i.hasNext() ;) {
            final Class factoryClass = (Class) i.next();
            vc.add(retrieveCallbackValuesArray(factoryClass));
        }
        return retrieveCombineStrategy(testClass).combine(vc);
    }
}
