/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit4;

import java.util.ArrayList;
import java.util.List;
import org.callbackparams.combine.reflect.CallbackRecordsFactory;
import org.junit.runner.Description;
import org.junit.runner.notification.RunNotifier;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.ParentRunner;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.InitializationError;

/**
 * The very simpliest use-case for callback-params is when the parameters are
 * self-contained, i.e. they have at least one regular JUnit4 test-method
 * (takes zero arguments) which can perform its test all on its own without
 * input from any containing callback-params test-method. For this case it
 * could be over-kill to have the parameters go the detour
 * through some customized CallbackParams test-class. However, by using this
 * test-runner class for the enum that provides these parameters it is possible
 * to run the self-contained test-methods without writing any additional code.
 * <br/><br/>
 * It is not absolutly mandatory for the test-class to be an enum. - The
 * important thing is that it has a static method "values()" that returns the
 * test-instances. I.e. for an enum-class the test-instances would then turn out
 * to be the enum-constants.
 * <br/><br/>
 * The implementation is pretty much the simpliest possible by taking advantage
 * of features available in the runner-class BlockJUnit4ClassRunner. The
 * downside of this is that it demands JUnit version 4.5 or later ...
 *
 * @author Henrik Kaipe
 */
public class EnumRunner extends ParentRunner<BlockJUnit4ClassRunner> {

    private final List<BlockJUnit4ClassRunner> children;
    
    public EnumRunner(Class<? extends Enum> testClass)
    throws InitializationError {
        super(testClass);
        Object[] enumValues = CallbackRecordsFactory.getInstance()
                .retrieveCallbackValuesArray(testClass);
        this.children = new ArrayList<BlockJUnit4ClassRunner>(enumValues.length);
        for (Object testInstance : enumValues) {
            this.children.add(
                    new SingleEnumValueRunner(testClass, testInstance));
        }
    }

    @Override
    protected List<BlockJUnit4ClassRunner> getChildren() {
        return this.children;
    }

    @Override
    protected Description describeChild(BlockJUnit4ClassRunner child) {
        return child.getDescription();
    }

    @Override
    protected void runChild(
            BlockJUnit4ClassRunner child, RunNotifier notifier) {
        child.run(notifier);
    }

    private static class SingleEnumValueRunner
    extends BlockJUnit4ClassRunner {
        
        private final Object testValue;

        public SingleEnumValueRunner(Class<?> testClass, Object testValue)
        throws InitializationError {
            super(testClass);
            this.testValue = testValue;
        }

        @Override
        protected Object createTest() {
            return testValue;
        }

        @Override
        protected String testName(FrameworkMethod method) {
            return super.testName(method) + '[' + testValue + ']';
        }

        @Override
        /**
         * Overridden to make sure that no constructor validations are performed.
         */
        protected void collectInitializationErrors(List<Throwable> errors) {
            validateInstanceMethods(errors);
        }
    }
}
